"""Test stuff on bit strings."""

from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng
from pycommons.types import check_int_range, type_error, type_name_of

from moptipy.algorithms.so.fitness import Fitness
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.execution import Execution
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_problem import MOProblem
from moptipy.api.objective import Objective
from moptipy.api.operators import Op0, Op1, Op1WithStepSize, Op2
from moptipy.examples.bitstrings.ising1d import Ising1d
from moptipy.examples.bitstrings.leadingones import LeadingOnes
from moptipy.examples.bitstrings.onemax import OneMax
from moptipy.examples.bitstrings.zeromax import ZeroMax
from moptipy.mo.problem.weighted_sum import WeightedSum
from moptipy.operators.bitstrings.op0_random import Op0Random
from moptipy.spaces.bitstrings import BitStrings
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.fitness import validate_fitness
from moptipy.tests.mo_algorithm import validate_mo_algorithm
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1
from moptipy.tests.op1_with_step_size import validate_op1_with_step_size
from moptipy.tests.op2 import validate_op2
from moptipy.utils.nputils import array_to_str


def dimensions_for_tests() -> Iterable[int]:
    """
    Get a sequence of dimensions for tests.

    :returns: the sequence of integers
    """
    r = default_rng()
    bs: list[int] = [1, 2, 3, 4, 5, 10, 16, 100,
                     int(r.integers(20, 50)), int(r.integers(200, 300))]
    r.shuffle(cast(list, bs))
    return bs


def bitstrings_for_tests() -> Iterable[BitStrings]:
    """
    Get a sequence of bit strings for tests.

    :returns: the sequence of BitStrings
    """
    return [BitStrings(i) for i in dimensions_for_tests()]


def random_bit_string(random: Generator, x: np.ndarray) -> np.ndarray:
    """
    Randomize a bit string.

    :param random: the random number generator
    :param x: the bit string
    :returns: the array
    """
    ri = random.integers
    for i in range(len(x)):  # pylint: disable=C0200
        x[i] = ri(2) <= 0
    return x


def validate_op0_on_1_bitstrings(
        op0: Op0 | Callable[[BitStrings], Op0],
        search_space: BitStrings,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the unary operator on one bit strings instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid": random_bit_string,
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_bitstrings(
        op0: Op0 | Callable[[BitStrings], Op0],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the unary operator on several BitStrings instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for bst in bitstrings_for_tests():
        validate_op0_on_1_bitstrings(op0, bst,
                                     number_of_samples, min_unique_samples)


def validate_op1_on_1_bitstrings(
        op1: Op1 | Callable[[BitStrings], Op1],
        search_space: BitStrings,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the unary operator on one BitStrings instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid": random_bit_string,
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_bitstrings(
        op1: Op1 | Callable[[BitStrings], Op1],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the unary operator on several BitStrings instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for bst in bitstrings_for_tests():
        validate_op1_on_1_bitstrings(op1, bst,
                                     number_of_samples, min_unique_samples)


def validate_op1_with_step_size_on_1_bitstrings(
        op1: Op1WithStepSize | Callable[[BitStrings], Op1WithStepSize],
        search_space: BitStrings,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, BitStrings], int] | None = None,
        step_sizes: Iterable[float] | Callable[
            [BitStrings], Iterable[float]] = (),
        get_step_size: Callable[[
            BitStrings, np.ndarray, np.ndarray,
        ], float | None] | None = None) -> None:
    """
    Validate the step-sized unary operator on one `BitStrings` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param step_sizes: the step sizes to test
    :param get_step_size: try to get the step size from two space elements
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid": random_bit_string,
        "step_sizes": step_sizes(search_space) if callable(step_sizes)
        else step_sizes,
        "get_step_size": get_step_size,
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1_with_step_size(**args)


def validate_op1_with_step_size_on_bitstrings(
        op1: Op1WithStepSize | Callable[[BitStrings], Op1WithStepSize],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, BitStrings], int] | None = None,
        step_sizes: Iterable[float] | Callable[
            [BitStrings], Iterable[float]] = (),
        get_step_size: Callable[[
            BitStrings, np.ndarray, np.ndarray,
        ], float | None] | None = None) -> None:
    """
    Validate the unary operator on several `BitStrings` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param step_sizes: the step sizes to test
    :param get_step_size: try to get the step size from two space elements
    """
    for bs in bitstrings_for_tests():
        validate_op1_with_step_size_on_1_bitstrings(
            op1, bs, number_of_samples, min_unique_samples, step_sizes,
            get_step_size)


def validate_op2_on_1_bitstrings(
        op2: Op2 | Callable[[BitStrings], Op2],
        search_space: BitStrings,
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the binary operator on one BitStrings instance.

    :param op2: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op2": op2(search_space) if callable(op2) else op2,
        "search_space": search_space,
        "make_search_space_element_valid": random_bit_string,
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op2(**args)


def validate_op2_on_bitstrings(
        op2: Op2 | Callable[[BitStrings], Op2],
        number_of_samples: int | None = None,
        min_unique_samples:
        int | Callable[[int, BitStrings], int] | None
        = None) -> None:
    """
    Validate the binary operator on several BitStrings instances.

    :param op2: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    for bst in bitstrings_for_tests():
        validate_op2_on_1_bitstrings(op2, bst,
                                     number_of_samples, min_unique_samples)


def validate_algorithm_on_bitstrings(
        objective: Objective | Callable[[int], Objective],
        algorithm: Algorithm | Callable[
            [BitStrings, Objective], Algorithm],
        dimension: int = 5, max_fes: int = 100,
        required_result: int | Callable[
            [int, int], int] | None = None,
        uses_all_fes_if_goal_not_reached: bool = True,
        post: Callable[[Algorithm, int], Any] | None = None) -> None:
    """
    Check the validity of a black-box algorithm on a bit strings problem.

    :param algorithm: the algorithm or algorithm factory
    :param objective: the objective function or function factory
    :param dimension: the dimension of the problem
    :param max_fes: the maximum number of FEs
    :param required_result: the optional required result quality
    :param uses_all_fes_if_goal_not_reached: will the algorithm use all FEs
        unless it reaches the goal?
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if not (isinstance(objective, Objective) or callable(objective)):
        raise type_error(objective, "objective", Objective, True)
    check_int_range(dimension, "dimension", 1, 1_000_000)
    if (post is not None) and (not callable(post)):
        raise type_error(post, "post", None, call=True)

    if callable(objective):
        objective = objective(dimension)
    if not isinstance(objective, Objective):
        raise type_error(objective, "result of callable 'objective'",
                         Objective)
    bs: Final[BitStrings] = BitStrings(dimension)
    if callable(algorithm):
        algorithm = algorithm(bs, objective)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "result of callable 'algorithm'",
                         Algorithm)

    goal: Final[int | None] = required_result(max_fes, dimension) \
        if callable(required_result) else required_result

    validate_algorithm(
        algorithm=algorithm, solution_space=bs, objective=objective,
        max_fes=max_fes, required_result=goal,
        uses_all_fes_if_goal_not_reached=uses_all_fes_if_goal_not_reached,
        post=post)


def validate_algorithm_on_onemax(
        algorithm: Algorithm | Callable[
            [BitStrings, Objective], Algorithm],
        post: Callable[[Algorithm, int], Any] | None = None) -> None:
    """
    Check the validity of a black-box algorithm on OneMax.

    :param algorithm: the algorithm or algorithm factory
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    """
    max_fes: Final[int] = 100
    for i in dimensions_for_tests():
        rr: int = 1 if i < 3 else (1 + max(1, i // 2, i - int(max_fes ** 0.5)))
        validate_algorithm_on_bitstrings(
            objective=OneMax,
            algorithm=algorithm,
            dimension=i,
            max_fes=max_fes,
            required_result=rr,
            post=post)


def validate_algorithm_on_leadingones(
        algorithm: Algorithm | Callable[
            [BitStrings, Objective], Algorithm],
        post: Callable[[Algorithm, int], Any] | None = None) -> None:
    """
    Check the validity of a black-box algorithm on LeadingOnes.

    :param algorithm: the algorithm or algorithm factory
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    """
    max_fes: Final[int] = 100
    for i in dimensions_for_tests():
        rr: int
        if i < 3:
            rr = 1
        elif max_fes > (10 * (i ** 1.5)):
            rr = i - 1
        else:
            rr = i
        validate_algorithm_on_bitstrings(
            objective=LeadingOnes,
            algorithm=algorithm,
            dimension=i,
            max_fes=int(1.4 * max_fes),
            required_result=rr,
            post=post)


def validate_mo_algorithm_on_bitstrings(
        problem: MOProblem | Callable[[int], MOProblem],
        algorithm: MOAlgorithm | Callable[
            [BitStrings, MOProblem], MOAlgorithm],
        dimension: int = 5,
        max_fes: int = 100) -> None:
    """
    Check a black-box multi-objective algorithm on a bit strings problem.

    :param algorithm: the algorithm or algorithm factory
    :param problem: the multi-objective optimization problem or factory
    :param dimension: the dimension of the problem
    :param max_fes: the maximum number of FEs
    """
    if not (isinstance(algorithm, MOAlgorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", MOAlgorithm, True)
    if not (isinstance(problem, MOProblem) or callable(problem)):
        raise type_error(problem, "problem", MOProblem, True)
    check_int_range(dimension, "dimension", 1, 1_000_000)

    if callable(problem):
        problem = problem(dimension)
    if not isinstance(problem, MOProblem):
        raise type_error(problem, "result of callable 'problem'",
                         MOProblem)
    bs: Final[BitStrings] = BitStrings(dimension)
    if callable(algorithm):
        algorithm = algorithm(bs, problem)
    if not isinstance(algorithm, MOAlgorithm):
        raise type_error(algorithm, "result of callable 'algorithm'",
                         MOAlgorithm)

    validate_mo_algorithm(algorithm=algorithm,
                          solution_space=bs,
                          problem=problem,
                          max_fes=max_fes)


def validate_mo_algorithm_on_2_bitstring_problems(
        algorithm: MOAlgorithm | Callable[
            [BitStrings, MOProblem], MOAlgorithm]) -> None:
    """
    Check the validity of a black-box algorithm on OneMax and ZeroMax.

    :param algorithm: the algorithm or algorithm factory
    """
    max_fes: Final[int] = 100
    random: Final[Generator] = default_rng()
    for i in dimensions_for_tests():
        weights: list[int | float] = [float(random.uniform(0.01, 10)),
                                      float(random.uniform(0.01, 10))] \
            if random.integers(2) <= 0 else \
            [1 + int(random.integers(1 << random.integers(40))),
             1 + int(random.integers(1 << random.integers(40)))]
        validate_mo_algorithm_on_bitstrings(
            problem=WeightedSum([OneMax(i), ZeroMax(i)], weights),
            algorithm=algorithm,
            dimension=i,
            max_fes=max_fes)


def validate_mo_algorithm_on_3_bitstring_problems(
        algorithm: MOAlgorithm | Callable[
            [BitStrings, MOProblem], MOAlgorithm]) -> None:
    """
    Check the validity of an algorithm on OneMax, ZeroMax, and Ising1d.

    :param algorithm: the algorithm or algorithm factory
    """
    max_fes: Final[int] = 100
    random: Final[Generator] = default_rng()
    for i in dimensions_for_tests():
        weights: list[int | float] = [float(random.uniform(0.01, 10)),
                                      float(random.uniform(0.01, 10)),
                                      float(random.uniform(0.01, 10))] \
            if random.integers(2) <= 0 else \
            [1 + int(random.integers(1 << random.integers(40))),
             1 + int(random.integers(1 << random.integers(40))),
             1 + int(random.integers(1 << random.integers(40)))]
        validate_mo_algorithm_on_bitstrings(
            problem=WeightedSum([OneMax(i), ZeroMax(i), Ising1d(i)],
                                weights),
            algorithm=algorithm,
            dimension=i,
            max_fes=max_fes)


def verify_algorithms_equivalent(
        algorithms: Iterable[Callable[[BitStrings, Objective], Algorithm]],
        max_fes: int | None = None) \
        -> None:
    """
    Verify that a set of algorithms performs identical steps.

    :param algorithms: the sequence of algorithms
    :param max_fes: the maximum number of FEs
    """
    if not isinstance(algorithms, Iterable):
        raise type_error(algorithms, "algorithms", Iterable)
    if (max_fes is not None) and (not isinstance(max_fes, int)):
        raise type_error(max_fes, "max_fes", int)

    random: Final[Generator] = default_rng()
    dim: Final[int] = int(random.integers(4, 16))
    space: Final[BitStrings] = BitStrings(dim)
    steps: Final[int] = int(random.integers(100, 1000)) \
        if max_fes is None else max_fes
    if steps <= 0:
        raise ValueError(f"invalid maximum FEs={steps} for max_fes={max_fes}")
    choice: Final[int] = int(random.integers(3))
    f: Final[Objective] = \
        LeadingOnes(dim) if choice <= 0 else \
        OneMax(dim) if choice <= 1 else \
        Ising1d(dim)
    evaluate: Final[Callable] = f.evaluate  # noqa
    seed: Final[int] = int(random.integers(1 << 62))

    result1: Final[list[bool]] = []
    result2: Final[list[bool]] = []
    first: bool = True
    first_name: str = ""
    do_fes: int = -1
    do_res: int | float = -1
    index: int = -1
    for algo in algorithms:
        index += 1
        if not callable(algo):
            raise type_error(algo, f"algorithms[{index}] for {f}", call=True)
        algorithm: Algorithm = check_algorithm(algo(space, f))
        current_name: str = str(algorithm)
        if first:
            first_name = current_name
            result = result1
        else:
            result = result2
            result.clear()

        def ff(x, rr=(result, )) -> int:
            nonlocal evaluate
            rres = evaluate(x)
            rr[0].extend(x)
            return rres

        ex = Execution()
        ex.set_algorithm(algorithm)
        ex.set_solution_space(space)
        f.evaluate = ff  # type: ignore
        ex.set_objective(f)
        ex.set_rand_seed(seed)
        ex.set_max_fes(steps)
        with ex.execute() as p:
            cf = p.get_consumed_fes()
            if not (0 < cf <= steps):
                raise ValueError(f"{current_name} consumed {cf} FS for "
                                 f"{steps} max FEs on {f} for seed {seed}.")
            if first:
                do_fes = cf
            elif do_fes != cf:
                raise ValueError(f"{current_name} consumed {cf} FEs but "
                                 f"{first_name} consumed {do_fes} FEs on "
                                 f"{f} for seed {seed}.")
            res = p.get_best_f()
            if not (0 <= res <= dim):
                raise ValueError(f"{current_name} got {res} as objective "
                                 f"value on {f} for seed {seed}.")
            if first:
                do_res = res
            elif do_res != res:
                raise ValueError(
                    f"{current_name} got {res} as objective value on {f} but "
                    f"{first_name} got {do_res} for seed {seed}.")
            if len(result) != (cf * dim):
                raise ValueError(
                    f"len(result) == {len(result)}, but should be {cf * dim} "
                    f"for {current_name} for seed {seed} on {f}.")
        if (not first) and (result1 != result2):
            raise ValueError(
                f"{current_name} produced different steps than {first_name} "
                f"on {f} for seed {seed}: is "
                f"{array_to_str(np.array(result2))}"
                f" but should be {array_to_str(np.array(result1))}.")

        first = False


def validate_fitness_on_bitstrings(
        fitness: Fitness | Callable[[Objective], Fitness],
        class_needed: str | type = Fitness,
        prepare_objective: Callable[[Objective], Objective] = lambda x: x) \
        -> None:
    """
    Validate a fitness assignment process on bit strings.

    :param fitness: the fitness assignment process, or a callable creating it
    :param class_needed: the required class
    :param prepare_objective: prepare the objective function
    """
    if (not isinstance(fitness, Fitness)) and (not callable(fitness)):
        raise type_error(fitness, "fitness", Fitness, call=True)
    if not isinstance(class_needed, str | type):
        raise type_error(class_needed, "class_needed", (str, type))
    if not callable(prepare_objective):
        raise type_error(prepare_objective, "prepare_objective", call=True)

    random: Final[Generator] = default_rng()
    sizes: set[int] = set()
    while len(sizes) < 4:
        sizes.add(int(random.integers(2, 10)))
    op0: Op0Random = Op0Random()
    for s in sizes:
        space: BitStrings = BitStrings(s)
        f: Objective = OneMax(s) if random.integers(2) <= 0 else LeadingOnes(s)
        f2 = prepare_objective(f)
        if not isinstance(f2, Objective):
            raise type_error(f2, f"prepare_objective({f})", Objective)
        del f
        if callable(fitness):
            ff = fitness(f2)
            if not isinstance(ff, Fitness):
                raise type_error(ff, f"fitness({f2})", Fitness)
            if isinstance(class_needed, str):
                name = type_name_of(ff)
                if name != class_needed:
                    raise TypeError(f"fitness assignment process should be "
                                    f"{class_needed!r}, but is {name!r}.")
            elif not isinstance(ff, class_needed):
                raise type_error(ff, f"fitness({f2})", class_needed)
        else:
            ff = fitness
        validate_fitness(ff, f2, space, op0)



"""
Tournament selection without replacement in the tournaments.

For each slot in the destination, a tournament with
:attr:`~moptipy.algorithms.modules.selections.\
tournament_without_repl.TournamentWithoutReplacement.size`
randomly chosen participating solutions is conducted. The solution with
the best fitness wins and is copied to the destination.

The tournaments are without replacement, following the algorithm by Goldberg,
Korg, and Deb. This means that a solution can only take part in another
tournament after all other solutions have joined one. In other words, we are
drawing solutions without replacement. This means that if we have `m`
solutions and want to select `n` from them by conducting tournaments of size
`s`, each solution will take part in at least `floor(s*n / m)` tournaments and
in at most `ceil(s*n / m)` ones.

We implement this drawing of unique random indices as a partial Fisher-Yates
shuffle. The indices used to choose the tournament contestants from are in an
array forming a permutation. Initially, `unused=m` indices. Whenever we draw
one new index from the permutation, we do so from a random position from
`0..unused-1`, swap it to position `unused-1` and decrement `unused` by one.
Once `unused` reaches `0`, we reset it to `m`. This is different from the
originally proposed implementation of tournament selection without repetition
in that it does not first permutate all the indices. It will be better in
cases where only a few small tournaments are conducted, e.g., during mating
selection. It will be slower when many large tournaments are conducted, e.g.,
during survival selection.

Tournament selection with replacement is implemented in
:mod:`moptipy.algorithms.modules.selections.tournament_with_repl`.

1. David Edward Goldberg and Bradley Korb and Kalyanmoy Deb. Messy Genetic
   Algorithms: Motivation, Analysis, and First Results. *Complex Systems*
   3(5):493-530. 1989.
   https://wpmedia.wolfram.com/uploads/sites/13/2018/02/03-5-5.pdf
2. Kumara Sastry and David Edward Goldberg. Modeling Tournament Selection with
   Replacement using Apparent Added Noise. In Lee Spector, Erik D. Goodman,
   Annie Wu, William Benjamin Langdon, and Hans-Michael Voigt, eds.,
   *Proceedings of the 3rd Annual Conference on Genetic and Evolutionary
   Computation (GECCO'01)*, July 7-11, 2001, San Francisco, CA, USA, page 781.
   San Francisco, CA, United States: Morgan Kaufmann Publishers Inc.
   ISBN: 978-1-55860-774-3. https://dl.acm.org/doi/pdf/10.5555/2955239.2955378
3. Peter J. B. Hancock. An Empirical Comparison of Selection Methods in
   Evolutionary Algorithms. In Terence Claus Fogarty, editor, *Selected Papers
   from the AISB Workshop on Evolutionary Computing (AISB EC'94),* April
   11-13, 1994, Leeds, UK, volume 865 of Lecture Notes in Computer Science,
   pages 80-94, Berlin/Heidelberg, Germany: Springer, ISBN: 978-3-540-58483-4.
   https://dx.doi.org/10.1007/3-540-58483-8_7. Conference organized by the
   Society for the Study of Artificial Intelligence and Simulation of
   Behaviour (AISB).
4. Uday Kumar Chakraborty and Kalyanmoy Deb and Mandira Chakraborty. Analysis
   of Selection Algorithms: A Markov Chain Approach. *Evolutionary
   Computation,* 4(2):133-167. Summer 1996. Cambridge, MA, USA: MIT Press.
   doi:10.1162/evco.1996.4.2.133.
   https://dl.acm.org/doi/pdf/10.1162/evco.1996.4.2.133
5. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
6. Sir Ronald Aylmer Fisher and Frank Yates. *Statistical Tables for
   Biological, Agricultural and Medical Research.* Sixth Edition, March 1963.
   London, UK: Oliver & Boyd. ISBN: 0-02-844720-4. https://digital.library.\
adelaide.edu.au/dspace/bitstream/2440/10701/1/stat_tab.pdf
"""

from math import inf
from typing import Any, Callable, Final

from numpy import empty, ndarray
from numpy.random import Generator
from pycommons.types import check_int_range

from moptipy.algorithms.modules.selection import FitnessRecord, Selection
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_INT, fill_in_canonical_permutation


# start book
class TournamentWithoutReplacement(Selection):
    """Tournament selection without replacement in the tournament."""

# end book
    def __init__(self, size: int = 2) -> None:
        """
        Create the tournament selection method.

        :param size: the size of the tournaments
        """
        super().__init__()
        #: the tournament size
        self.size: Final[int] = check_int_range(size, "tournament size", 1)
        #: the cache for the array
        self.__perm: ndarray = empty(0, DEFAULT_INT)

# start book
    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:
        """
        Perform deterministic best selection without replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        size: Final[int] = self.size  # the tournament size
        m: Final[int] = len(source)  # number of elements to select from
        perm: ndarray = self.__perm  # get the base permutation
        if len(perm) != m:  # -book
            self.__perm = perm = empty(m, DEFAULT_INT)  # -book
            fill_in_canonical_permutation(perm)  # -book
        r0i = random.integers  # fast call to random int from 0..(i-1)
        u: int = m  # the number u of available (unused) indices = m
        for _ in range(n):  # conduct n tournaments
            best: FitnessRecord | None = None  # best competitor
            best_fitness: int | float = inf  # best fitness, initial infinite
            for __ in range(size):  # perform one tournament
                if u == 0:  # if we have used all indices in perm
                    u = m  # then we again at the end
                i = r0i(u)  # get random integer in 0..u-1
                u = u - 1  # decrease number of unused indices
                chosen = perm[i]  # get the index of the chosen element
                perm[u], perm[i] = chosen, perm[u]  # swap to the end
                rec = source[chosen]  # get contestant record
                rec_fitness = rec.fitness  # get its fitness
                if rec_fitness <= best_fitness:  # if better or equal...
                    best = rec  # ... rec becomes the new best record
                    best_fitness = rec_fitness  # and remember fitness
            dest(best)  # at end of the tournament, send best to dest
        # end book

    def __str__(self):
        """
        Get the name of the tournament selection algorithm.

        :return: the name of the tournament selection algorithm
        """
        return f"tour{self.size}"

    def initialize(self) -> None:
        """Initialize this selection algorithm."""
        super().initialize()
        fill_in_canonical_permutation(self.__perm)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("size", self.size)



"""
Choose the `n` best elements from the source array.

"Best" selection is the standard way to perform survival selection in a
(mu+lambda) Evolutionary Algorithm (:class:`~moptipy.algorithms.so.ea.EA`).
It basically sorts the input array based on fitness and then copies the first
`n` elements to the output array. The algorithm is not random but
deterministic, i.e., it will always select each of the `n` best records
exactly once.

This selection is also often called "mu+lambda," "mu,lambda," or "top-n"
selection. It is similar to the selection scheme in Eshelman's CHC algorithm.
A randomized version of this algorithm is called "truncation selection" in the
paper of Blickle and Thiele.

1. Peter J. B. Hancock. An Empirical Comparison of Selection Methods in
   Evolutionary Algorithms. In Terence Claus Fogarty, editor, *Selected Papers
   from the AISB Workshop on Evolutionary Computing (AISB EC'94),* April
   11-13, 1994, Leeds, UK, volume 865 of Lecture Notes in Computer Science,
   pages 80-94, Berlin/Heidelberg, Germany: Springer, ISBN: 978-3-540-58483-4.
   https://dx.doi.org/10.1007/3-540-58483-8_7. Conference organized by the
   Society for the Study of Artificial Intelligence and Simulation of
   Behaviour (AISB).
2. Larry J.Eshelman. The CHC Adaptive Search Algorithm: How to Have Safe
   Search When Engaging in Nontraditional Genetic Recombination. In Gregory
   J. E. Rawlins, editor, Foundations of Genetic Algorithms, volume 1, 1991,
   pages 265-283. San Francisco, CA, USA: Morgan Kaufmann.
   https://doi.org/10.1016/B978-0-08-050684-5.50020-3
3. Frank Hoffmeister and Thomas Bäck. Genetic Algorithms and Evolution
   Strategies: Similarities and Differences. In Hans-Paul Schwefel and
   Reinhard Männer, *Proceedings of the International Conference on Parallel
   Problem Solving from Nature (PPSN I),* October 1-3, 1990, Dortmund,
   Germany, volume 496 of Lecture Notes in Computer Science, pages 455-469,
   Berlin/Heidelberg, Germany: Springer. ISBN: 978-3-540-54148-6.
   https://doi.org/10.1007/BFb0029787.
4. Uday Kumar Chakraborty and Kalyanmoy Deb and Mandira Chakraborty. Analysis
   of Selection Algorithms: A Markov Chain Approach. *Evolutionary
   Computation,* 4(2):133-167. Summer 1996. Cambridge, MA, USA: MIT Press.
   doi:10.1162/evco.1996.4.2.133.
   https://dl.acm.org/doi/pdf/10.1162/evco.1996.4.2.133
5. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
"""

from typing import Any, Callable

from numpy.random import Generator

from moptipy.algorithms.modules.selection import FitnessRecord, Selection


# start book
class Best(Selection):
    """The best selection: select each of the best `n` elements once."""

    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:  # pylint: disable=W0613
        """
        Perform deterministic best selection without replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        source.sort()  # sort by fitness, best solutions come first
        for i in range(n):  # select the n first=best solutions
            dest(source[i])  # by sending them to dest
# end book

    def __str__(self):
        """
        Get the name of the best selection algorithm.

        :return: the name of the best selection algorithm
        """
        return "best"



"""
Search operators for permutations with repetitions.

These operators can be used in conjunction with the space
:class:`~moptipy.spaces.permutations.Permutations`.

- Module :mod:`~moptipy.operators.permutations.op0_shuffle` provides a nullary
  operator sampling an entirely random permutation.
- Module :mod:`~moptipy.operators.permutations.op1_insert1` provides a unary
  operator that extracts one element and inserts it elsewhere.
- Module :mod:`~moptipy.operators.permutations.op1_swap2` provides a unary
  operator which swaps exactly two different elements of a permutation.
- Module :mod:`~moptipy.operators.permutations.op1_swapn` provides a unary
  operator which tries to find a cyclic move whose length would be
  approximately binomially distributed, i.e., which may swap two elements with
  probability 0.5, three elements with probability 0.25, 4 elements with
  probability 0.125, and so on. It does so by performing a sequence of swaps
  but does not check whether one swap undoes a previous one.
- Module :mod:`~moptipy.operators.permutations.op1_swap_exactly_n` offers a
  unary operator with a step size (see
  :class:`~moptipy.api.operators.Op1WithStepSize`). A step size of `0.0` means
  that it will swap two elements, a step size of `1.0` means that it tries to
  perform the largest-possible modification. If it is applied to permutations
  where each element occurs once, this means that all elements change their
  position. If applied to permutations with repetitions, things are more
  complex and some moves may be impossible, in which case a best effort is
  attempted.
- Module :mod:`~moptipy.operators.permutations.op1_swap_try_n` is similar to
  the operator from :mod:`~moptipy.operators.permutations.op1_swap_exactly_n`,
  but invests less effort into reaching the prescribed number of modifications
  exactly. Instead, it will accept making only fewer swaps more easily.
- Module :mod:`~moptipy.operators.permutations.op2_gap` offers an operator
  that tries to build a new permutation by appending not-yet-appended elements
  from both input permutations, alternating between them randomly.
- Module :mod:`~moptipy.operators.permutations.op2_ox2` implements a slightly
  modified version of the order-based crossover operator for permutations.
"""



"""
A unary operator flipping a pre-defined number of bits.

This is a unary operator with step size, i.e., an instance of
:class:`~moptipy.api.operators.Op1WithStepSize`. As such, it also receives
a parameter `step_size` when it is applied, which is from the closed range
`[0.0, 1.0]`. If `step_size=0`, then exactly 1 bit will be flipped. If
`step_size=1`, then all bits will be flipped. For all values of
`0<step_size<1`, we use the function
:func:`~moptipy.operators.tools.exponential_step_size` to extrapolate the
number of bits to flip.

Unary operators like this are often used in (1+1)-EAs or even in
state-of-the-art EAs such as the Self-Adjusting (1+(lambda,lambda)) GA.

1. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   doi: https://doi.org/10.1109/TEVC.2022.3191698,
   https://arxiv.org/pdf/2112.00229.pdf.
2. Eduardo Carvalho Pinto and Carola Doerr. *Towards a More Practice-Aware
   Runtime Analysis of Evolutionary Algorithms,* 2018,
   arXiv:1812.00493v1 [cs.NE] 3 Dec 2018. https://arxiv.org/abs/1812.00493.
"""
from typing import Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1WithStepSize
from moptipy.operators.tools import exponential_step_size


class Op1FlipM(Op1WithStepSize):
    """A unary search operation that flips a specified number of `m` bits."""

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray,
            step_size: float = 0.0) -> None:
        """
        Copy `x` into `dest` and flip exactly `m` bits.

        `step_size=0.0` will flip exactly `1` bit, `step_size=1.0` will flip
        all `n` bits. All other values are extrapolated by function
        :func:`~moptipy.operators.tools.exponential_step_size`. In between
        the two extremes, an exponential scaling
        (:func:`~moptipy.operators.tools.exponential_step_size`) is performed,
        meaning that for `n=10` bits, a step-size of `0.2` means flipping
        two bits, `0.4` means flipping three bits, `0.6` means flipping four
        bits, `0.7` means flipping five bits, `0.9` means flipping eight bits,
        `0.95` means flipping nine bits, and `1.0` means flipping all bits.
        In other words, a larger portion of the `step_size` range corresponds
        to making small changes while the large changes are all condensed at
        the higher end of the scale.

        :param self: the self pointer
        :param random: the random number generator
        :param dest: the destination array to receive the new point
        :param x: the existing point in the search space
        :param step_size: the number of bits to flip

        >>> op1 = Op1FlipM()
        >>> from numpy.random import default_rng as drg
        >>> rand = drg()
        >>> import numpy as npx
        >>> src = npx.zeros(10, bool)
        >>> dst = npx.zeros(10, bool)
        >>> for ss in [0.0, 0.2, 0.4, 0.6, 0.7, 0.8, 0.85, 0.9, 0.95, 1.0]:
        ...   op1.op1(rand, dst, src, ss)
        ...   print(sum(dst != src))
        1
        2
        3
        4
        5
        6
        7
        8
        9
        10
        """
        np.copyto(dest, x)  # copy source to destination
        n: Final[int] = len(dest)  # get the number of bits
        flips: Final[np.ndarray] = random.choice(  # choose the bits
            n, exponential_step_size(step_size, 1, n), False)
        dest[flips] ^= True  # flip the selected bits via xor

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :return: "flipm"
        """
        return "flipm"



"""
A meta-algorithm for model-assisted optimization using SciPy's Interpolation.

This algorithm wraps another numerical optimization algorithm `A` and proceeds
in two stages.
First, it samples and evaluates a set of initial points during the warmup
phase. These points are directly sampled by `A` on the original process, the
meta-algorithm just collects them.

Then, in the second stage, for each iteration, a model is constructed from
all previously sampled and evaluated points. The model is used for
interpolating the actual objective function.
In each step, the inner algorithm `A` is applied to this model. It strictly
works on the model and does not invoke the original objective. Instead, we
maintain the best point that `A` has sampled on the model based on the modeled
objective function. This best point is then evaluated on the actual objective
function. Together with its actual objective value, it is added to the set of
evaluated points. In the next step, a new model will be constructed based on
all the points we have now. This model is then the basis for the next
"simulated" run of `A`, and so on.

Thus, in the second stage, each execution of `A` on the model problem yields
one new point that is actually evaluated. The new point is used to create a
better model, and so on. If the models reflect the actual objective function
well, this may allow us to achieve better overall solution qualities or to
reduce the number of actual objective function evaluations to reach certain
goals.

However, this only works if a) we will not do too many actual objective
function evaluations (FEs) overall, as the memory requirement grows
quadratically with the number of FEs and b) if the dimensionality of the
problem is not too high, as the number of points needed to create a reasonably
accurate model rises with the dimensions of the search space.
"""

from math import inf
from typing import Callable, Final

import numpy as np
from pycommons.types import check_int_range, type_error
from scipy.interpolate import RBFInterpolator  # type: ignore
from scipy.special import comb  # type: ignore

from moptipy.algorithms.so.vector.surrogate._processes import (
    _SurrogateApply,
    _SurrogateWarmup,
)
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.process import Process
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_FLOAT
from moptipy.utils.strings import num_to_str_for_name, sanitize_names

#: the permitted RBF kernels
_RBF_KERNELS: Final[dict[str, str]] = {
    "linear": "l",
    "thin_plate_spline": "tps",
    "cubic": "c",
    "quintic": "q",
}


class RBFInterpolation(Algorithm):
    """
    A meta algorithm using an RBF-interpolation based surrogate model.

    This algorithm uses :class:`scipy.interpolate.RBFInterpolator` as
    interpolator surrogate model.
    """

    def __init__(self,
                 space: VectorSpace,
                 inner: Algorithm,
                 fes_for_warmup: int,
                 fes_per_interpolation: int,
                 kernel: str = "thin_plate_spline",
                 degree: int = 2,
                 name="RBF") -> None:
        """
        Create an interpolation-based surrogate algorithm.

        :param name: the base name of this algorithm
        :param inner: the algorithm to be applied in the inner optimization
            loop
        :param fes_for_warmup: the number of objective function evaluations to
            be spent on the initial warmup period
        :param fes_per_interpolation: the number of FEs to be performed
            for each interpolation run
        """
        super().__init__()

        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        if not isinstance(kernel, str):
            raise type_error(kernel, "kernel", str)
        if kernel not in _RBF_KERNELS:
            raise ValueError(
                f"kernel={kernel!r} not permitted, must be one "
                f"of {_RBF_KERNELS.keys()}.")

        degree = check_int_range(degree, "degree", -1, 20)
        dimensions = check_int_range(
            space.dimension, "space.dimensions", 1, 10_000)
        fes_for_warmup = check_int_range(
            fes_for_warmup, "fes_for_warmup", 1, 1_000_000_000)
        if degree >= 0:
            min_points: Final[int] \
                = int(comb(degree + dimensions, dimensions, exact=True))
            fes_for_warmup = max(fes_for_warmup, min_points)

        #: the vector space
        self._space = space
        #: the inner algorithm
        self.__inner = check_algorithm(inner)
        #: the warmup FEs
        self.__warmup_fes = fes_for_warmup
        #: the FEs per interpolation run
        self.__interpolation_fes = check_int_range(
            fes_per_interpolation, "fes_per_interpolation", 1,
            1_000_000_000_000)
        #: the name of this surrogate assisted
        self.__name = sanitize_names((
            f"{name}{_RBF_KERNELS[kernel]}{num_to_str_for_name(degree)}",
            str(fes_for_warmup), str(fes_per_interpolation), str(inner)))
        #: the kernel name
        self.__kernel: Final[str] = kernel
        #: the degree
        self.__degree: Final[int] = degree

    def __str__(self) -> str:
        """
        Get the name of this surrogate-assisted algorithm.

        :returns: the name of this surrogate assisted algorithm
        """
        return self.__name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this surrogate-assisted algorithm.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("warmupFEs", self.__warmup_fes)
        logger.key_value("interpolationFEs", self.__interpolation_fes)
        logger.key_value("kernel", self.__kernel)
        logger.key_value("degree", self.__degree)
        with logger.scope("inner") as inner:
            self.__inner.log_parameters_to(inner)
        with logger.scope("space") as space:
            self._space.log_parameters_to(space)

    def solve(self, process: Process) -> None:
        """
        Apply the surrogate-assisted optimization method to the given process.

        :param process: the process to solve
        """
        # fast calls
        should_terminate: Final[Callable[[], bool]] = process.should_terminate
        inner: Final[Callable[[Process], None]] = self.__inner.solve
        evaluate: Final[Callable[[np.ndarray], int | float]] = \
            process.evaluate
        init: Final[Callable[[], None]] = self.__inner.initialize
        uniform: Final[Callable[[
            np.ndarray, np.ndarray, int], np.ndarray]] = \
            process.get_random().uniform
        empty: Final[Callable[[int, np.dtype], np.ndarray]] \
            = np.empty

        # constants
        lb: Final[np.ndarray] = self._space.lower_bound
        ub: Final[np.ndarray] = self._space.upper_bound
        dim: Final[int] = self._space.dimension
        dtype: Final[np.dtype] = self._space.dtype
        run_fes: Final[int] = self.__interpolation_fes
        kernel: Final[str] = self.__kernel
        degree: Final[int] = self.__degree

# the containers for the points that we have sampled
        x: Final[list[np.ndarray]] = []  # the points that were sampled so far
        z: Final[list[int | float]] = []  # their objective values

# Perform the initial warm-up process. Here, the inner algorithm will directly
# be applied to the original problem. All the points that it samples are
# collected and will later be used to construct the model.
        with _SurrogateWarmup(process, self.__warmup_fes,
                              x.append, z.append) as p2:
            p2._fes_left = p2.max_fes  # type: ignore # store the budget
            p2._terminated = False  # type: ignore # not terminated yet
            init()  # initialize the inner algorithm
            inner(p2)  # apply the inner algorithm to the real model
        del p2

# Now, we have collected self.__warmup_fes points from the search space.

        if should_terminate():
            return

# We can now perform the optimization on the model. The model is constructed
# based on all points in the search space that were sampled and evaluated with
# the actual objective function. In each iteration, we apply the inner
# algorithm to the model from scratch. After it has terminated, then take the
# best point it found (based on the modeled objective function) and evaluate
# it with the actual objective function. This point and its objective value
# are then added to the internal list and used, together with all previous
# points, to construct the model for the next iteration.
        model: Final[_SurrogateApply] = _SurrogateApply(process, run_fes)

        while True:
            while True:
                # We always begin by building the surrogate model anew.
                # However, this may sometimes fail. Maybe a parameter matrix
                # becomes singular or whatever.
                try:
                    f: Callable[[np.ndarray], np.ndarray] = \
                        RBFInterpolator(np.array(x, dtype=DEFAULT_FLOAT),
                                        np.array(z, dtype=DEFAULT_FLOAT),
                                        kernel=kernel,
                                        degree=degree)
                    break  # success: quit innermost loop
                except:  # noqa # pylint: disable=[W0702]
                    # If we get here, the model construction has failed.
                    # This means that the points that we have collected are
                    # somehow insufficient.
                    # We try to fix this by sampling one additional point
                    # uniformly at random and evaluate it.
                    # If this does not exhaust the FEs that we have, we can
                    # then try again.
                    tmp = uniform(lb, ub, dim)
                    x.append(tmp)  # add random point to list of points
                    z.append(process.evaluate(tmp))  # and its objective value
                    if process.should_terminate():  # did we exhaust budget?
                        return  # yes ... so we return

            model._fes_left = run_fes  # assign the budget for the run
            model._terminated = False  # the run has not terminated
            model._evaluate = f  # forward evaluation to the model
            model._best_f = inf  # no best-so-far solution exists yet
            init()  # initialize the inner algorithm
            inner(model)  # apply the inner algorithm to the model
            tmp = empty(dim, dtype)  # allocate holder for result
            model.get_copy_of_best_x(tmp)  # get best solution
            z2 = evaluate(tmp)  # evaluate it on the actual problem
            if should_terminate():  # should we quit?
                return  # yes, so we return
            x.append(tmp)  # add the best solution to the list of points
            z.append(z2)  # and also add the objective value
            del f  # dispose old model



"""
The EAFEA-A is hybrid of the (1+1)FEA and the (1+1)EA with Solution Transfer.

The solution is transferred from the FEA branch to the EA branch if its
H-value is 1, i.e., if it represents a completely new objective value.

1. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, Sarah Louise
   Thomson, and Thomas Weise. Addressing the Traveling Salesperson Problem
   with Frequency Fitness Assignment and Hybrid Algorithms. *Soft Computing.*
   2024. https://dx.doi.org/10.1007/s00500-024-09718-8
"""
from collections import Counter
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


class EAFEAA(Algorithm1):
    """An implementation of the EAFEA-A."""

    def __init__(self, op0: Op0, op1: Op1, log_h_tbl: bool = False) -> None:
        """
        Create the EAFEA-A.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param log_h_tbl: should we log the H table?
        """
        super().__init__("eafeaA", op0, op1)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        #: True if we should log the H table, False otherwise
        self.log_h_tbl: Final[bool] = log_h_tbl

    def solve(self, process: Process) -> None:
        """
        Apply the EAFEA-A to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        x_c = process.create()  # record for current solution of the EA
        x_d = process.create()  # record for current solution of the FEA
        x_n = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        xcopy: Final[Callable] = process.copy  # copy(dest, source)
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, x_c)  # Create 1 solution randomly and
        y_c: int | float = evaluate(x_c) + ofs  # evaluate it.
        xcopy(x_d, x_c)
        y_d: int | float = y_c
        use_ffa: bool = True

        while not should_terminate():  # Until we need to quit...
            use_ffa = not use_ffa  # toggle use of FFA
            op1(random, x_n, x_d if use_ffa else x_c)
            y_n: int | float = evaluate(x_n) + ofs

            if use_ffa:  # the FEA branch
                h[y_n] += 1  # type: ignore  # Increase the frequency
                h[y_d] += 1  # type: ignore  # of new_f and cur_f.
                h_n = h[y_n]  # type: ignore
                if h_n <= h[y_d]:  # type: ignore
                    y_d = y_n  # Store its objective value.
                    x_d, x_n = x_n, x_d  # Swap best and new.
                    if h_n <= 1:  # if solution is new, then transfer it to
                        xcopy(x_c, x_d)  # the EA/RLS branch
            elif y_n <= y_c:  # the EA/RLS branch
                y_c = y_n
                x_c, x_n = x_n, x_c

        if not self.log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[y_c] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[y_c] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table



"""A nullary operator filling a :mod:`~moptipy.spaces.ordered_choices`."""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op0
from moptipy.spaces.ordered_choices import OrderedChoices


class Op0ChooseAndShuffle(Op0):
    """Randomly initialize :mod:`~moptipy.spaces.ordered_choices` elements."""

    def __init__(self, space: OrderedChoices):
        """
        Initialize this shuffle operation: use blueprint from space.

        :param space: the search space
        """
        super().__init__()
        if not isinstance(space, OrderedChoices):
            raise type_error(space, "space", OrderedChoices)
        #: the internal blueprint for filling permutations
        self.__blueprint: Final[np.ndarray] = space.blueprint
        #: the internal choices
        self.__choices: Final[Callable[[int], tuple[int, ...]]] \
            = space.choices.__getitem__

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the destination with random choices and shuffle it.

        :param random: the random number generator
        :param dest: the destination array
        """
        np.copyto(dest, self.__blueprint)  # Copy blueprint to dest.
        random.shuffle(dest)  # Shuffle destination array randomly.

        # Now we replace the elements in the shuffled destination array
        # with random selections.
        ri: Final[Callable[[int], int]] = random.integers  # fast call
        choices: Final[Callable[[int], tuple[int, ...]]] = self.__choices
        for i, e in enumerate(dest):
            source = choices(e)
            dest[i] = source[ri(len(source))]

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "chooseAndShuffle"
        """
        return "chooseAndShuffle"



"""
Select `n` records at random without replacement.

This selection scheme is the standard mating selection scheme in an
Evolutionary Algorithm.
"""

from typing import Any, Callable, Final

from numpy.random import Generator

from moptipy.algorithms.modules.selection import FitnessRecord, Selection


# start book
class RandomWithoutReplacement(Selection):
    """Select random elements without replacement."""

    def select(self, source: list[FitnessRecord],
               dest: Callable[[FitnessRecord], Any],
               n: int, random: Generator) -> None:
        """
        Select `n` random elements from `source` without replacement.

        :param source: the list with the records to select from
        :param dest: the destination collector to invoke for each selected
            record
        :param n: the number of records to select
        :param random: the random number generator
        """
        m: Final[int] = len(source)
        if n == 1:  # handle n=1 exactly as in (mu+lambda) EA
            dest(source[random.integers(m)])  # pick 1 solution randomly
        elif n == 2:  # handle n=2 exactly as in (mu+lambda) EA
            ri = random.integers  # fast call
            a = b = ri(m)  # get first random index
            while a == b:  # find a second, different random index
                b = ri(m)
            dest(source[a])  # send first solution to dest
            dest(source[b])  # send second solution to dest
        else:  # handle other cases: n ints from 0..m-1 w/o replacement
            for i in random.choice(m, n, False):  # get the ints
                dest(source[i])  # send randomly chosen records to dest
# end book

    def __str__(self):
        """
        Get the name of the random choice without replacement selection.

        :return: the name of the random choice without replacement selection
            algorithm
        """
        return "rndNoRep"



"""
Permutations of a selection (or ordered choices) as strings.

Imagine the following situation: We have a rectangle (two-dimensional) and
want to place `n` other, smaller rectangles into it. Let each rectangle have
a unique ID from `1` to `n`. We can use a permutation of `1..n` to represent
the order in which we insert the small rectangles. However, maybe we want to
also be able to rotate a rectangle by 90° and then place it. So additionally
to the insertion order, we may want to remember whether the rectangle is
placed as-is or rotated. We could do this with a "permutation of choices",
where we can place either `i` or `-i` for `i` from `1..n`. A negative value
could mean "insert rotated by 90°" whereas a positive value means "insert as
is". So we have `n` (disjoint) choices, each of which with two options. From
each choice, we can pick one value. Then the order in which the values appear
marks the insertion order. So this is basically a "super space" of
permutations, as it deals both with the order of the elements and their values
(resulting from the selected choice).

A string consists of `n` elements. There also are `n` so-called "selections,"
each of which being a set offering a choice from  different values. Any two
selections must either be entirely disjoint or equal. The final string must
contain one value from each selection.

Let's say that we have three selections, e.g., `[1, 2, 3]`, `[4, 5]`, and
`[6]`. Then the "selection permutations" space contains, e.g., the string
`[4, 3, 6]` or `[2, 5, 6]` -- one value from each selection. It does not
contain `[1, 3, 5]`, though, because that string has two values from the first
selection.

This space is a super space of the :mod:`~moptipy.spaces.permutations`, i.e.,
the space of permutations with repetitions. Sometimes (check
:meth:`OrderedChoices.is_compatible_with_permutations`), the search
operators defined in package :mod:`~moptipy.operators.permutations` can also
be applied to the elements of our space here, although they may not be able
to explore the space in-depth (as they will not alter the choices and only
permute the chosen elements).
"""
from collections import Counter
from math import factorial
from typing import Final, Iterable

import numpy as np
from pycommons.types import type_error

from moptipy.spaces.intspace import IntSpace
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import array_to_str

#: the different choices
KEY_CHOICES: Final[str] = "choices"


class OrderedChoices(IntSpace):
    """Permutations of selections, stored as :class:`numpy.ndarray`."""

    def __init__(self, selections: Iterable[Iterable[int]]) -> None:
        """
        Create the space of permutations of selections.

        :param selections: an iterable of selections
        :raises TypeError: if one of the parameters has the wrong type
        :raises ValueError: if the parameters have the wrong value
        """
        if not isinstance(selections, Iterable):
            raise type_error(selections, "selections", Iterable)

        sets: Final[dict[int, tuple[int, ...]]] = {}

        min_v: int = 0
        max_v: int = 0
        total: int = 0
        counts: Counter[tuple[int, ...]] = Counter()
        blueprint: Final[list[int]] = []

        for i, sel in enumerate(selections):
            total += 1
            if not isinstance(sel, Iterable):
                raise type_error(sel, f"selections[{i}]", Iterable)
            sel_lst = tuple(sorted(sel))
            len_sel = len(sel_lst)
            if len_sel <= 0:
                raise ValueError(f"empty selection at index {i}: {sel}.")
            sel_set = set(sel_lst)
            if len(sel_set) != len_sel:
                raise ValueError(f"invalid selection {sel} at index {i} "
                                 f"contains duplicate element")
            blueprint.append(sel_lst[0])
            # build the selection map
            for e in sel_lst:
                if not isinstance(e, int):
                    raise type_error(e, f"selections[{i}]={sel}", int)
                if e in sets:
                    lst_found = sets[e]
                    if lst_found != sel_lst:
                        raise ValueError(
                            f"value {e} appears both in selection {sel_lst} "
                            f"(at permutation index {i}) and in selection "
                            f"{lst_found}!")
                    sel_lst = lst_found
                    continue  # if any sets[e] != sel_lst, we get error anyway
                sets[e] = sel_lst  # remember value
            counts[sel_lst] += 1

            # update the value range
            if total <= 1:
                min_v = sel_lst[0]
                max_v = sel_lst[-1]
            else:
                min_v = min(min_v, sel_lst[0])
                max_v = max(max_v, sel_lst[-1])

        if total <= 0:
            raise ValueError(
                "there must be at least one selection, "
                f"but got {selections}.")

        # creating super class
        super().__init__(total, min_v, max_v)

        #: the selector map
        self.choices: Final[dict[int, tuple[int, ...]]] = sets
        #: how often does each element choice appear?
        self.__counts: Final[dict[tuple[int, ...], int]] = {
            k: counts[k] for k in sorted(counts.keys())}
        blueprint.sort()

        #: The blueprint array, i.e., an ordered array holding the smallest
        #: value possible for each choice.
        self.blueprint: Final[np.ndarray] = np.array(
            blueprint, dtype=self.dtype)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)

        strings: Final[list[str]] = []
        for sel, count in self.__counts.items():
            s = CSV_SEPARATOR.join(map(str, sel))
            strings.append(s if count <= 1 else (s + "*" + str(count)))
        logger.key_value(KEY_CHOICES, "/".join(strings))

    def create(self) -> np.ndarray:
        r"""
        Create an ordered choice, equal to :attr:`~OrderedChoices.blueprint`.

        :return: the ordered choices, an instance of :class:`numpy.ndarray`.

        >>> perm = OrderedChoices([[1, 3], [2, 4], [1, 3], [7, 5]])
        >>> x = perm.create()
        >>> print(perm.to_str(x))
        1;1;2;5
        """
        return self.blueprint.copy()  # Create copy of the blueprint.

    def validate(self, x: np.ndarray) -> None:
        """
        Validate a permutation of selections.

        :param x: the integer string
        :raises TypeError: if the string is not an :class:`numpy.ndarray`.
        :raises ValueError: if the element is not a valid permutation of the
            given choices
        """
        super().validate(x)

        usage: Final[Counter[tuple[int, ...]]] = Counter()
        dct: Final[dict[int, tuple[int, ...]]] = self.choices
        for j, i in enumerate(x):
            if i not in dct:
                raise ValueError(
                    f"invalid element {i} encountered at index"
                    f" {j} of string {array_to_str(x)}.")
            usage[dct[i]] += 1
        counts: Final[dict[tuple[int, ...], int]] = self.__counts
        for tup, cnt in usage.items():
            expected = counts[tup]
            if expected != cnt:
                raise ValueError(
                    f"encountered value from {tup} exactly {cnt} times "
                    f"instead of the expected {expected} times in {x}")

    def n_points(self) -> int:
        """
        Get the number of possible different permutations of the choices.

        :return: [factorial(dimension) / Product(factorial(count(e))
            for all e)] * Product(len(e)*count(e) for all e)
        """
        # get the basic permutation numbers now multiply them with the choices
        size = factorial(self.dimension)
        for lst, cnt in self.__counts.items():
            size = size // factorial(cnt)
            size = size * (len(lst) ** cnt)
        return size

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "selperm{n}", where {n} is the length
        """
        return f"selperm{len(self.blueprint)}"

    def is_compatible_with_permutations(self) -> bool:
        """
        Check whether for compatibility with permutations with repetitions.

        Or, in other words, check whether the operators in package
        :mod:`~moptipy.operators.permutations` can safely be applied for
        elements of this space here.

        Permutations with repetitions are permutations where each element
        occurs exactly a given number of times. Our implementation of this
        space (:mod:`~moptipy.spaces.permutations`) ensures that there are
        at least two different elements. The unary and binary search
        operators defined in package :mod:`~moptipy.operators.permutations`
        rely on this fact. While these operators cannot explore the depth
        of the permutations-of-selections space here, they can be "compatible"
        to this space: Any element of the permutation-of-selections space is,
        by definition, a permutation with repetitions, as it contains one
        concrete manifestation per choice. Applying, for instance, a
        :mod:`~moptipy.operators.permutations.op1_swap2` operation to it,
        which swaps two different elements, still yields a valid and different
        permutation-of-selections. However, since the operators in
        :mod:`~moptipy.operators.permutations` always enforce that the
        resulting point is different from their input and *only* permute the
        elements, this can only work if we have at least two disjoint choices
        in our space definition. The function here checks this.

        :returns: `True` if and only if the operators in package
            :mod:`~moptipy.operators.permutations` can safely be applied to
            elements of this space
        """
        return len(self.__counts) > 1

    @staticmethod
    def signed_permutations(n: int) -> "OrderedChoices":
        """
        Create a space for signed permutations with values `1..n`.

        You would be much better off using
        :mod:`~moptipy.spaces.signed_permutations` instead of this space for
        signed permutations, though.

        :param n: the range of the values
        :returns: the permutations space

        >>> perm = OrderedChoices.signed_permutations(3)
        >>> perm.validate(perm.blueprint)
        >>> print(perm.blueprint)
        [-3 -2 -1]
        >>> print(perm.n_points())  # 3! * (2 ** 3) = 6 * 8 = 48
        48
        >>> perm = OrderedChoices.signed_permutations(4)
        >>> perm.validate(perm.blueprint)
        >>> print(perm.blueprint)
        [-4 -3 -2 -1]
        >>> print(perm.n_points())  # 4! * (2 ** 4) = 24 * 16 = 384
        384
        """
        return OrderedChoices([-i, i] for i in range(1, n + 1))



"""A base class for bitstring-based problems."""

from typing import Final

from pycommons.types import check_int_range

from moptipy.api.objective import Objective
from moptipy.spaces.bitstrings import BitStrings
from moptipy.utils.logger import KeyValueLogSection


class BitStringProblem(Objective):
    """
    A base class for problems defined over bit strings.

    This base class has a set of default behaviors. It has an attribute
    :attr:`n` denoting the lengths of the accepted bit strings. Its
    :meth:`lower_bound` returns `0` and its :meth:`upper_bound` returns
    :attr:`n`. :meth:`is_always_integer` returns `True`.
    """

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the bitstring objective function.

        :param n: the dimension of the problem
        """
        super().__init__()
        #: the length of the bit strings
        self.n: Final[int] = check_int_range(n, "n", 1)

    def lower_bound(self) -> int:
        """
        Get the lower bound of the bit string based problem.

        :return: 0

        >>> print(BitStringProblem(10).lower_bound())
        0
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the bit string based problem.

        :return: the length of the bit string

        >>> print(BitStringProblem(7).upper_bound())
        7
        """
        return self.n

    def is_always_integer(self) -> bool:
        """
        Return `True` if the `evaluate` function always returns an `int`.

        This pre-defined function for bit-string based problems will always
        return `True`.

        :retval True: always
        """
        return True

    def space(self) -> BitStrings:
        """
        Create a proper search space for this problem.

        :returns: an instance of
            :class:`~moptipy.spaces.bitstrings.BitStrings` for bit strings of
            length :attr:`n`
        """
        return BitStrings(self.n)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         BitStringProblem(5).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: BitStringProblem'
        >>> text[3]
        'lowerBound: 0'
        >>> text[4]
        'upperBound: 5'
        >>> text[5]
        'n: 5'
        >>> len(text)
        7
        """
        super().log_parameters_to(logger)
        logger.key_value("n", self.n)



"""Providing a process without explicit logging with a single space."""
from typing import Final, cast

from moptipy.api._process_base import _TIME_IN_NS, _ns_to_ms, _ProcessBase
from moptipy.api.logging import (
    PROGRESS_CURRENT_F,
    PROGRESS_FES,
    PROGRESS_TIME_MILLIS,
    SECTION_PROGRESS,
)
from moptipy.utils.logger import Logger


class _ProcessNoSS(_ProcessBase):
    """
    An internal class process implementation.

    This class implements a stand-alone process without explicit logging where
    the search and solution space are the same.
    """

    def evaluate(self, x) -> float | int:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and "
                                 "the algorithm knows it.")
            return self._current_best_f

        result: Final[int | float] = self._f(x)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if result < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = result
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (result <= self._end_f)
            self._copy_y(self._current_best_y, x)

        if do_term:
            self.terminate()

        return result

    def register(self, x, f: int | float) -> None:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and "
                                 "the algorithm knows it.")
            return

        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        if f < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = f
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (f <= self._end_f)
            self._copy_y(self._current_best_y, x)

        if do_term:
            self.terminate()

    def __str__(self) -> str:
        return "ProcessWithoutSearchSpace"


def _write_log(log: list[list[int | float]],
               start_time: int,
               logger: Logger) -> None:
    """
    Write the log to a logger.

    :param log: the log
    :param start_time: the start time
    :param logger: the logger
    """
    if len(log) > 0:
        with logger.csv(SECTION_PROGRESS,
                        [PROGRESS_FES,
                         PROGRESS_TIME_MILLIS,
                         PROGRESS_CURRENT_F]) as csv:
            for row in log:
                csv.row([row[0], _ns_to_ms(cast(int, row[1]) - start_time),
                         row[2]])



"""
An objective function counting the number of zeros in a bit string.

This problem exists mainly for testing purposes as counterpart of
:class:`~moptipy.examples.bitstrings.onemax.OneMax`.
"""

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def zeromax(x: np.ndarray) -> int:
    """
    Get the length of a string minus the number of zeros in it.

    :param x: the np array
    :return: the length of the string minus the number of zeros, i.e., the
        number of ones

    >>> print(zeromax(np.array([True, True, False, False, False])))
    2
    >>> print(zeromax(np.array([True, False, True, False, False])))
    2
    >>> print(zeromax(np.array([False, True,  True, False, False])))
    2
    >>> print(zeromax(np.array([True, True, True, True, True])))
    5
    >>> print(zeromax(np.array([False, True, True, True, True])))
    4
    >>> print(zeromax(np.array([False, False, False, False, False])))
    0
    """
    return int(x.sum())


class ZeroMax(BitStringProblem):
    """Maximize the number of zeros in a bit string."""

    def __init__(self, n: int) -> None:
        """
        Initialize the zeromax objective function.

        :param n: the dimension of the problem

        >>> print(ZeroMax(2).n)
        2
        >>> print(ZeroMax(4).evaluate(np.array([True, True, False, True])))
        3
        """
        super().__init__(n)
        self.evaluate = zeromax  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the zeromax objective function.

        :return: `zeromax_` + length of string

        >>> print(ZeroMax(13))
        zeromax_13
        """
        return f"zeromax_{self.n}"
